""" This module contains some experiments I intend to put into my
    thesis chapter on Fast elastic registration

    and also to the article in Image Processing

    Jan Kybic, April 2001
    $Id: felregexp.py,v 1.2 2002/01/21 16:33:19 jkybic Exp $
    """

import shelve,time
import Image
from Numeric import *
import bigtools
import regultrasound
import bigfelreg
import bigoptimize
import bigregister
import bigsplines
import Gnuplot
import RandomArray,MA
import bigfract

# ----------------------------------------------------------------------

datadir='data_felreg/'
imsize=(576,768) # raw image size
#vsize=(0.5*array(imsize)).astype(Int)
vsize=imsize  # size of images to show
scalefactor=0.25

def readDistant():
    """ returns two ultrasound images """
    if 1:     # read big images from the raw data
        imgn1='images_chus/Im-4C-Clip0004-masked/4C-Clip0004-768x576-0_mask.raw'
        imgn2='images_chus/Im-4C-Clip0004-masked/4C-Clip0004-768x576-10_mask.raw'
        img1=regultrasound.readRawMatlabBytes(imgn1,imsize)
        img2=regultrasound.readRawMatlabBytes(imgn2,imsize)
        # resize images and save them
        bigtools.ArrayToImage(img1).resize((int(scalefactor*imsize[1]),int(scalefactor*imsize[0]))).save(datadir+'Im-4C-Clip0004-0s.tiff')
        bigtools.ArrayToImage(img2).resize((int(scalefactor*imsize[1]),int(scalefactor*imsize[0]))).save(datadir+'Im-4C-Clip0004-10s.tiff')
    else: # read smaller images
        imgn1=datadir+'Im-4C-Clip0004-0s.tiff'
        imgn2=datadir+'Im-4C-Clip0004-10s.tiff'
        img1=bigtools.ImageToArray(Image.open(imgn1))
        img2=bigtools.ImageToArray(Image.open(imgn2))
    return img1,img2

def registerDistant():
    """ This experiment is intended to show how a registration fails
    without landmarks and succeeds with landmarks

    It takes about 8s with quarter size images and
    92s with full-sized ones (for crate=(64,64))
    """

    img1,img2=readDistant()
    #v1=bigtools.view(img1,title='Image 1',size=vsize)
    #v2=bigtools.view(img2,title='Image 2',size=vsize)
    #imgdif=abs(bigtools.int_array(img1)-img2)
    #difscale=(0,200)
    v3=bigtools.view(bigtools.showDif(img1,img2),
                     title='difference before',size=vsize)
    bigtools.printtofile(bigtools.showDif(img1,img2),
                         datadir+'regdist-difb.tiff')
    # for smaller images crate=(16,16) is appropriate
    par=bigtools.Parameters({'verbose':5, 'crate':(64,64),
                             'xtol':0.5, 'degc':3, 'degf':1 ,
                             'alternatered':1, 'grmask':0,
                             'testimg':img2, 'refimg':img1})
    wp=bigfelreg.MultigridableWarpingProblemBase(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    mrs.solveMR()
    x=mrs.getProblem().getXstate() # get the deformation parameters
    img2w=bigfelreg.WarpableImageS(img2,par=par).setDefS(x).getWarped()
    v4=bigtools.view(img2w,title='Warped image 2',size=vsize)
    bigtools.printtofile(img2w,datadir+'regdist-img2w.tiff')
    v5=bigtools.view(bigtools.showDif(img1,img2w),
                     title='difference after',size=vsize)
    bigtools.printtofile(bigtools.showDif(img1,img2w),
                         datadir+'regdist-difa.tiff')
    grid=bigtools.makeGrid(img2.shape)
    dg=bigfelreg.WarpableImageS(grid,par=par).setDefS(x).getWarped()
    v6=bigtools.view(dg,title='deformed grid',size=vsize)
    bigtools.printtofile(dg,datadir+'regdist-gridw.tiff')
    raw_input('Press any key')

def registerDistantLandmarks():
    """ This experiment is intended to show how a registration
    succeeds with landmarks """

    img1,img2=readDistant()
    # give landmarks in the original coordinates, (y,x) ordering
    lnd1=array(((324,490),(409,412),(348,306),(174,347)),Float) # landmark(s) in image 1
    lnd2=array(((277,493),(421,376),(261,268),(163,371)),Float) # landmark(s) in image 2
    sclf=float(img1.shape[0])/imsize[0] # calculate scaling
    lnd1*=sclf ; lnd2*=sclf
    # 10 works well for small images
    # lndf=repeat([2.0],(lnd1.shape[0],))
    lndf=array((1,0.1,0.1,0.1),Float)/sclf/sclf
    v1=bigtools.view(bigtools.showImgWithLandmarks(img1,lnd1,lndsize=4),
                     title='Image 1',size=vsize)
    bigtools.printtofile(bigtools.showImgWithLandmarks(img1,lnd1,lndsize=4),
                         datadir+'regdistl-img1.tiff')
    v2=bigtools.view(bigtools.showImgWithLandmarks(img2,lnd2,lndsize=4),
                     title='Image 2',size=vsize)
    bigtools.printtofile(bigtools.showImgWithLandmarks(img2,lnd2,lndsize=4),
                         datadir+'regdistl-img2.tiff')
    # raw_input('Press any key') ; return
    v3=bigtools.view(bigtools.showDif(img1,img2),
                      title='difference before',size=vsize)
    par=bigtools.Parameters({'verbose':6, 'crate':(16,16),
                             'xtol':0.5, 'degc':3, 'degf':1 ,
                             'alternatered':1, 'grmask':0,
                             'testimg':img2, 'refimg':img1,
                             'minlevel':0,
                             'springpos': lnd1, 'springtarget':lnd2,
                             'springfactor':lndf})
    wp=bigfelreg.MultigridableWarpingProblemSprings(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    mrs.solveMR()
    x=mrs.getProblem().getXstate() # get the deformation parameters
    img2w=bigfelreg.WarpableImageS(img2,par=par).setDefS(x).getWarped()
    v4=bigtools.view(img2w,title='Warped image 2',size=vsize)
    bigtools.printtofile(img2w,datadir+'regdistl-img2w.tiff')
    v5=bigtools.view(bigtools.showDif(img1,img2w),
                      title='difference after',size=vsize)
    bigtools.printtofile(bigtools.showDif(img1,img2w),
                         datadir+'regdistl-difa.tiff')
    grid=bigtools.makeGrid(img2.shape)
    dg=bigfelreg.WarpableImageS(grid,par=par).setDefS(x).getWarped()
    v6=bigtools.view(dg,title='deformed grid',size=vsize)
    bigtools.printtofile(dg,datadir+'regdistl-gridw.tiff')
    raw_input('Press any key')

def showRBFresult():
    """ Shows the result of pure Thin plate spline landmark registration
    with the same data as in registerDistantLandmarks()

    The landmark deformation is calculated by script
    work/matlabsrc/landmarks/showfelreglnd.m
    """
    img1,img2=readDistant()
    img2w=bigtools.ImageToArray(Image.open(datadir+'showfelreglnd.tiff'))
    v5=bigtools.view(bigtools.showDif(img1,img2w),
                     title='RBF difference after',size=vsize)
    bigtools.printtofile(bigtools.showDif(img1,img2w),
                         datadir+'rbfresult-difa.tiff')
    raw_input('Press any key')
    

# ----------------------------------------------------------------------

def showRegistration():
    """ This experiment shows how the criterion and the registration
    error evolve during time. """
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    asize=2*array(ima.shape)
    # verbosity may not be higher than 4 if we want nice graphs !
    # orig settings (used in the article were 0.5 and 0.1 (xtollast)
    par=bigtools.Parameters({'xtol':0.5, 'xtollast': 0.1,
                             'verbose':4, 'crate':(32,32),
                             'maxlevel':5})
    wi=bigfelreg.WarpableImageS(ima,par=par)
    c=wi.getCAsArray() ; print "c.shape=",c.shape
    c[0,1:4,3:5]=10.0 ; c[1,5:8,5:7]=-15.0 
    c[1,:,:]+=4.0 # shift down 
    wi.setCFromArray(c)
    imw=wi.getWarped()
    mask=loadMask()
    #mask=bigregister.calculateThresholdedMask(imw,threshold=40,degs=3)
    bigtools.printtofile(mask,datadir+'showreg-mask.tiff')
    #v0=bigtools.view(mask,title='mask',size=asize)
    # v1=bigtools.view(imw,title='Warped',size=asize)
    #v2=bigtools.view(bigtools.showDif(imw,ima),
    #                 title='Difference before',size=asize)
    bigtools.printtofile(bigtools.showDif(imw,ima),
                         datadir+'showreg-difb.tiff')
    reslist=[] ; reslisttime=[]
    par=bigtools.Parameters(par,default={'testimg':ima,'refimg':imw,
                          'degc':3, 'degf':1, 'alternatered':1,      
                          'trueg':wi.getDeformation(),'wimask':mask,
                          'reslist': reslist})
    part=bigtools.Parameters(par,
            override={'reslist':reslisttime, 'measuretime':1})
    wp=bigfelreg.MultigridableWarpingProblemWI(par=par)
    iniwi=wp.datapart.testw.getWarpingIndexDefS(defs=wi.getDefS(),par=par,
                                                mask=mask)
    print "initial wi=",iniwi
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    bigtools.timeop(lambda mrs=mrs: mrs.solveMR(),'solveMR')
    
    x=mrs.getProblem().getXstate()
    cf=mrs.getOptimizer().getX()
    #print cf
    print "cdif=", bigtools.maxelement(abs(c-cf))
    #wp.datapart.testw.setCFromArray(cf)
    print "initial wi=",iniwi
    print "wi=",wp.datapart.testw.getWarpingIndexDefS(defs=wi.getDefS(),
                                                      par=par,mask=mask)
    g0=wp.datapart.testw.getDeformation()
    g1=wi.getDeformation()
    rmask=repeat(reshape(mask,[1]+list(mask.shape)),array((2,)))
    g0=MA.MaskedArray(g0,mask=1-rmask)
    gdif=sqrt(add.reduce(((g0-g1)**2).filled(fill_value=0.0)))
    print 'max geom.error=',bigtools.maxelement(gdif)
    #v3a=bigtools.view(bigtools.showDif(gdif*20,imw),size=asize,title='Geometric error')
    gm=bigregister.calculateMask(ima,pct=5) # almost contours
    difimg=bigtools.showDif(gm,gdif)
    v3a=bigtools.view(difimg,size=asize,title='Geometric error')
    raw_input('Press any key')

    bigtools.printtofile(difimg,datadir+'showreg-geomerr.tiff')
    
    imaw=wi.setDefS(x).getWarped()
    v3=bigtools.view(bigtools.showDif(imw,imaw),
                     title='Difference after',size=asize)
    bigtools.printtofile(bigtools.showDif(imw,imaw),
                         datadir+'showreg-difa.tiff')
    # register again to measure time
    wp=bigfelreg.MultigridableWarpingProblemWI(par=part)
    mrs=bigoptimize.MROptimizerState(wp,par=part)
    bigtools.timeop(lambda mrs=mrs: mrs.solveMR(),'solveMR')
    f=shelve.open(datadir+"showreg-reslist.dat")
    f['reslist']=reslist ;     f['reslisttime']=reslisttime
    f.close()
    raw_input('Press any key')


# ----------------------------------------------------------------------

def showReslist():
    """ Shows a graph of the evolution of the registration in showRegistration
    The reslist is read from a shelved file. It is a list of
    (imgsize, crate,E,wi)
    """
    f=shelve.open(datadir+"showreg-reslist.dat")
    reslist=f['reslist']
    reslisttime=f['reslisttime']
    f.close()
    E=array([ x[2] for x in reslist],Float)
    wi=array([ x[3] for x in reslist],Float)
    maximgsize=array([ max(x[0]) for x in reslist],Float)
    maxcrate=array([ max(x[1]) for x in reslist],Float)
    nim=E.shape[0]  # number of elements
    #itn=arange(nim) # iteration (evaluation) number
    itime=array(reslisttime) # computer time
    itime-=itime[0]
    #g=Gnuplot.Gnuplot()
    #dE=Gnuplot.Data(itn,E,with='lines',title='criterion') 
    #dwi=Gnuplot.Data(itn,wi,with='lines',title='warping index')
    #g.plot(dE,dwi)
    g1=bigtools.plot(E,title='criterion',with='lines',xlabel='iteration number',
                     ylabel='mean SSD')
    g1("set terminal postscript 20")
    g1.hardcopy(datadir+"showreg-Eit.eps")
    g2=bigtools.plot(wi,title='warping index',with='lines',
                     xlabel='iteration number', ylabel='pixels')
    g2("set terminal postscript 20")
    g2.hardcopy(datadir+"showreg-wiit.eps")
    g3=bigtools.plot(maximgsize,title='imgsize',with='lines',
                     xlabel='iteration number', ylabel='pixels')
    g3("set terminal postscript 20")
    g3.hardcopy(datadir+"showreg-imgit.eps")
    g4=bigtools.plot(maximgsize.astype(Float)/maxcrate,
                     title='size/crate',with='lines',
                     xlabel='iteration number', ylabel='ncoefs')
    g4("set terminal postscript 20")
    g4.hardcopy(datadir+"showreg-ncit.eps")
    g11=bigtools.plot(itime,E,title='criterion',with='lines',
                      xlabel='time in seconds',
                     ylabel='mean SSD')
    g11("set terminal postscript 20")
    g11.hardcopy(datadir+"showreg-Etime.eps")
    g12=bigtools.plot(itime,wi,title='warping index',with='lines',
                     xlabel='time in seconds', ylabel='pixels')
    g12("set terminal postscript 20")
    g12.hardcopy(datadir+"showreg-witime.eps")
    g13=bigtools.plot(itime,maximgsize,title='imgsize',with='lines',
                     xlabel='time in seconds', ylabel='pixels')
    g13("set terminal postscript 20")
    g13.hardcopy(datadir+"showreg-imgime.eps")
    g14=bigtools.plot(itime,maximgsize.astype(Float)/maxcrate,
                     title='size/crate',with='lines',
                     xlabel='time in seconds', ylabel='ncoefs')
    g14("set terminal postscript 20")
    g14.hardcopy(datadir+"showreg-ncoeftime.eps")

    raw_input('Press any key')
    
# ----------------------------------------------------------------------

def showMask():
    """ Calculates a gradient mask for Lena and shows it """
    img=bigtools.ImageToArray(Image.open('lena.tiff'))
    mask=1-bigregister.calculateMask(img,pct=10)
    v1=bigtools.view(mask)
    bigtools.printtofile(mask,datadir+'lenamask.tiff')
    raw_input('Press any key')
    
# ----------------------------------------------------------------------

def saveMask():
    """ Saves thresholded mask for the MRI picture """
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    mask=bigregister.calculateThresholdedMask(ima,threshold=30,degs=0)
    bigtools.printtofile(mask,datadir+'mrithrmask.tiff')

def loadMask():
    """ Loads mask for the MRI picture from file """
    mask=bigtools.ImageToArray(Image.open(datadir+'mrithrmask_edited.tiff'))
    mask=mask.astype(UnsignedInt8)
    return mask

# ----------------------------------------------------------------------

def showSpaces():
    """ Shows how the warping index is affected when we go to
    coarser or finer grids """
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    asize=2*array(ima.shape)
    par=bigtools.Parameters({'xtol':0.1, 'xtollast': 0.01, 'degc':3,
                             'degf':3, 'verbose':6,
                             'crate':(256,256)})
    wi=bigfelreg.WarpableImageS(ima,par=par)
    mask=loadMask()
    defszero=wi.getDefS().copy() ; parzero=par
    # get the initial deformation
    c=wi.getCAsArray() ; print "c.shape=",c.shape
    # get the expanded version gradually perturbing
    w=wi
    RandomArray.seed(34323451,178935720) # for debugging, set seed
    for i in xrange(6):
        if i>0: w=w.getExpandedDef()
        c=w.getCAsArray() ; c+=RandomArray.normal(0.0,20.0/2**i,c.shape)
        w.setCFromArray(c)
    grid=bigtools.makeGrid(ima.shape)
    dg=bigfelreg.WarpableImageS(grid,par=w.par).setDefS(w.getDefS()).getWarped()
    v1=bigtools.view(dg,title='original deformed grid')
    bigtools.printtofile(dg,datadir+'showspaces-ordefgrid.tiff')
    wrdefs=w.getDefS().copy() # reference deformation
    wrpar=w.par
    imw=w.getWarped() # get the warped version of the image
    gr=w.getDeformation() # reference deformation
    #wi.setProjection(gr,mask=mask)
    wil=[]
    for i in xrange(6):
        if i>0:
            w=w.getReducedDef()           # reduce
            print "windexbefore=",w.getWarpingIndexDefS(wrdefs,wrpar,mask=mask)
            w.setProjection(gr,mask=mask) # improve fit
            print "windexafter=",w.getWarpingIndexDefS(wrdefs,wrpar,mask=mask)
            
        windex=w.getWarpingIndexDefS(wrdefs,wrpar,mask=mask)
        windexz=w.getWarpingIndexDefS(defszero,parzero,mask=mask)
        regpar=bigtools.Parameters(w.par,default={
            'testimg':ima, 'refimg':imw})
        wp=bigfelreg.MultigridableWarpingProblemBase(par=regpar)
        # wp.setXstate(w.getDefS()) # starting guess, maybe not fair
        windexbef=wp.datapart.testw.getWarpingIndexDefS(
            defs=wrdefs,par=wrpar,mask=mask)
        mrs=bigoptimize.MROptimizerState(wp,par=regpar)
        bigtools.timeop(lambda mrs=mrs: mrs.solveMR(),'solveMR')
        windexaft=wp.datapart.testw.getWarpingIndexDefS(
            defs=wrdefs,par=wrpar,mask=mask)
        if 1:
            x=mrs.getProblem().getXstate()
            imaw=wp.datapart.testw.getWarped()
            dg=bigfelreg.WarpableImageS(grid,par=w.par).setDefS(x).getWarped()
            v2=bigtools.view(dg,title='recovered grid')
            bigtools.printtofile(dg,datadir+'showspaces-recoveredgrid%d.tiff' % i)
#            v3=bigtools.view(bigtools.showDif(imw,imaw),
#                     title='difference after',size=vsize)
            v3=bigtools.view(imaw-imw,size=vsize,scale=(-200.0,200.0))
            print "Eaft=",bigtools.sumall(abs(imw-imaw)**2)
            bigtools.printtofile(imaw-imw,
                                 datadir+'showspaces-difafter%0d.tiff'%i,
                                 scale=(-200.0,200.0))
#            bigtools.printtofile(bigtools.showDif(imw,imaw),
#                                 datadir+'showspaces-difafter%0d.tiff'%i)

#            raw_input('Press any key')

        wil.append((windex,windexz,windexbef,windexaft,w.par.crate[0]))
        print "at level ",i,"csize=",w.getDefS().first().c.shape,\
              "crate=",w.par.crate," wi=",windex, " wiz=",windexz,\
              "windexbef=", windexbef," windexaft=", windexaft
    print wil
    f=shelve.open(datadir+"showspaces-windlist.dat")
    f['windlist']=wil ; 
    f.close()

def showSpacesRes():
    """ Shows a graph from the data saved by showSpaces """
    f=shelve.open(datadir+"showspaces-windlist.dat")
    wil=f['windlist'] ; 
    f.close()
    wind=array([ x[0] for x in wil],Float)
    wiaft=array([ x[3] for x in wil],Float)
    #cr=array([ x[4] for x in wil],Float)
    cr=arange(wind.shape[0])
    g=Gnuplot.Gnuplot()
    dind=Gnuplot.Data(cr,wind,with='linespoints',title='ideal') 
    daft=Gnuplot.Data(cr,wiaft,with='linespoints',title='recovered')
    g('set xlabel "knot spacing [pixels/knot]"')
    g('set xtics ("8" 0,"16" 1,"32" 2,"64" 3,"128" 4,"256" 5)')
    g('set ylabel "warping index [pixels]"')
    g.plot(dind,daft)
    g("set terminal postscript 20")
    g.hardcopy(datadir+'showspaces-wind.eps')
    raw_input('Press any key')
    
# ----------------------------------------------------------------------

def showStatDist():
    """ Generate many deformations, recover them, and compare warping
    results """
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    asize=2*array(ima.shape)
    # was 0.1 0.01
    par=bigtools.Parameters({'xtol':0.1, 'xtollast': 0.01, 'degc':3,
                             'degf':3, 'verbose':4,
                             'crate':(128,128)})
    wi=bigfelreg.WarpableImageS(ima,par=par)
    mask=loadMask()
    dref=wi.getDefS().copy()
    RandomArray.seed(34312351,1789372) # for debugging, set seed
    wires=[]
    for j in xrange(1000): # repeat experiments this many times
        # generate deformation
        w=bigfelreg.WarpableImageS(wi.imgs,defs=wi.getDefS(),par=par)
        for i in xrange(3):
            if i>0: w=w.getExpandedDef()
            c=w.getCAsArray()
            c+=RandomArray.normal(0,10.0/4**i,c.shape)
            w.setCFromArray(c)
        imw=w.getWarped() # get the warped version of the image
        #grid=bigtools.makeGrid(imw.shape)
        #wg=bigfelreg.WarpableImageS(grid,par=w.par).setDefS(w.getDefS())
        #dg=wg.getWarped()
        #v1=bigtools.view(dg,size=asize)
        #raw_input('Press any key')
        
        windexbef=w.getWarpingIndexDefS(dref,par=par,mask=mask)
        regpar=bigtools.Parameters(w.par,default={
            'testimg':ima, 'refimg':imw, 'crate': (32,32)})
        wp=bigfelreg.MultigridableWarpingProblemBase(par=regpar)
        mrs=bigoptimize.MROptimizerState(wp,par=regpar)
        bigtools.timeop(lambda mrs=mrs: mrs.solveMR(),'solveMR')
        windexaft=wp.datapart.testw.getWarpingIndexDefS(w.getDefS(),par=w.par,
                                                        mask=mask)
        wires.append((windexbef,windexaft))
        print "iteration ",j,"wdef=",windexbef,\
              "werr=",windexaft
        f=shelve.open(datadir+"showstatdist.dat")
        f['wires']=wires ; 
        f.close()

def showStatRes():
    """ present the results generated from showStatDist """
    f=shelve.open(datadir+"showstatdist.dat")
    wil=f['wires'] ; 
    f.close()
    print len(wil)
    wdef=array([ x[0] for x in wil],Float)
    werr=array([ x[1] for x in wil],Float)
    g0=Gnuplot.Gnuplot()
    d0=Gnuplot.Data(wdef,werr,with='points')

    maxr=max((bigtools.maxelement(wdef),bigtools.maxelement(werr)))
    #g0("set xrange [0:%f]" % maxr)
    #g0("set yrange [0:%f]" % maxr)
    #g0("set xrange [0:80]" )
    #g0("set yrange [0:50]" )
    g0('set xlabel "original wi [pixels]"') 
    g0('set ylabel "error wi [pixels]"')
    g0.plot(d0)
    g0("set terminal postscript 25")
    g0.hardcopy(datadir+'showstatres-orerr.eps')
    t1,h1=bigtools.calchist(wdef,minx=-0.5,h=0.1)
    g1=bigtools.plot(t1,h1,with='boxes',xlabel='original wi [pixels]',ylabel='count')
    g1("set terminal postscript 25")
    g1.hardcopy(datadir+'showstatres-horr.eps')
    #t2,h2=bigtools.calchist(werr,h=0.002)
    t2,h2=bigtools.calchist(werr,h=0.1,minx=-2,maxx=18)
    g2=bigtools.plot(t2,h2,with='boxes',xlabel='error wi [pixels]',ylabel='count')
    g2("set terminal postscript 25")
    g2.hardcopy(datadir+'showstatres-herr.eps')
    raw_input('Press any key')

# ----------------------------------------------------------------------

def showDegDep():
    """ Show the dependence of the registration time and precision
    on the spline orders (for both image and model) """
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    par=bigtools.Parameters({'xtol':0.5, 'xtollast': 0.1, 'degc':3,
                             'degf':3, 'verbose':4,
                             'crate':(256,256)})
    wi=bigfelreg.WarpableImageS(ima,par=par)
    mask=loadMask()
    # get the expanded version gradually perturbing
    w=wi
    RandomArray.seed(43123451,17835720) # for debugging, set seed
    for i in xrange(6):
        if i>0: w=w.getExpandedDef()
        c=w.getCAsArray() ; c+=RandomArray.normal(0.0,15.0/1.5**i,c.shape)
        w.setCFromArray(c)
    wrdefs=w.getDefS().copy() # reference deformation
    wrpar=w.par
    imw=w.getWarped() # get the warped version of the image
    resl=[]
    for i in xrange(1,4): # spline order i
        regpar=bigtools.Parameters(par,override={'degc':i,'crate':(64,64),
                                   'testimg':ima,'refimg':imw})
        wp=bigfelreg.MultigridableWarpingProblemBase(par=regpar)
        mrs=bigoptimize.MROptimizerState(wp,par=regpar)
        windex=wp.datapart.testw.getWarpingIndexDefS(
            defs=wrdefs,par=wrpar,mask=mask)
        t1=time.clock()
        mrs.solveMR()
        te=time.clock()-t1
        windexaft=wp.datapart.testw.getWarpingIndexDefS(
            defs=wrdefs,par=wrpar,mask=mask)
        print "degc=%d te=%g wibef=%f wi=%f" % (i,te,windex,windexaft)
        regpar=bigtools.Parameters(par,override={'degf':i,'crate':(64,64),
                                   'testimg':ima,'refimg':imw})
        wp=bigfelreg.MultigridableWarpingProblemBase(par=regpar)
        mrs=bigoptimize.MROptimizerState(wp,par=regpar)
        t1=time.clock()
        mrs.solveMR()
        tef=time.clock()-t1
        windexaftf=wp.datapart.testw.getWarpingIndexDefS(
            defs=wrdefs,par=wrpar,mask=mask)
        print "degf=%d te=%g wibef=%f wi=%f" % (i,tef,windex,windexaftf)
        resl.append((windex,i,windexaft,te,windexaftf,tef))
        f=shelve.open(datadir+"showdegdep.dat")
        f['resl']=resl ; 
        f.close()

def showDegDepRes():
    """ present results calculated in showDegDep """
    f=shelve.open(datadir+"showdegdep.dat")
    resl=f['resl'] ; 
    f.close()
    print "Initial wi=", resl[0][2]
    print "degf & degc & wi & time \\\\"
    for i in resl:
        print "3 & %d & %0.3f & %0.1f \\\\" % (i[1],i[2],i[3])
        print "%d & 3 & %0.3f & %0.1f \\\\" % (i[1],i[4],i[5])


# ----------------------------------------------------------------------

def showStartPInfl():
    """ Show influence of the starting point on the result """
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    asize=2*array(ima.shape)
    # verbosity may not be higher than 4 if we want nice graphs !
    # orig settings (used in the article were 0.5 and 0.1 (xtollast)
    par=bigtools.Parameters({'xtol':0.5, 'xtollast': 0.01,
                             'verbose':4, 'crate':(32,32),
                             'maxlevel':0 # only last level
                             })
    wi=bigfelreg.WarpableImageS(ima,par=par)
    cbase=wi.getCAsArray() ; print "c.shape=",cbase.shape
    cbase[0,1:4,3:5]=10.0 ; cbase[1,5:8,5:7]=-15.0 
    cbase[1,:,:]+=4.0 # shift down
    wi.setCFromArray(cbase)
    imw=wi.getWarped()
    mask=loadMask()
    regpar=bigtools.Parameters(par,default={
            'testimg':ima, 'refimg':imw})

    resl=[]
    for alpha in list(arange(0,1.01,0.05)):
        wp=bigfelreg.MultigridableWarpingProblemBase(par=regpar)
        wp.setX(cbase*alpha)
        windex=wp.datapart.testw.getWarpingIndexDefS(
            wi.getDefS(),par=par,mask=mask)
        mrs=bigoptimize.MROptimizerState(wp,par=regpar)
        bigtools.timeop(lambda mrs=mrs: mrs.solveMR(),'solveMR')
        windexaft=wp.datapart.testw.getWarpingIndexDefS(
            wi.getDefS(),par,mask=mask)
        print "alpha=",alpha," windex=",windex, " wiaft=",windexaft
        resl.append((alpha,windex,windexaft))
        f=shelve.open(datadir+"showStartPInfl.dat")
        f['resl']=resl ; 
        f.close()

def showStartPInflRes():
    """ presents results calculated by showStartPInfl """
    f=shelve.open(datadir+"showStartPInfl.dat")
    resl=f['resl'] ; 
    f.close()
    alph=array([ x[0] for x in resl],Float)
    windex=array([ x[1] for x in resl],Float)
    windexaft=array([ x[2] for x in resl],Float)
    #g=Gnuplot.Gnuplot()
    #dind=Gnuplot.Data(alph,windex,with='linespoints',title='initial') 
    #daft=Gnuplot.Data(alph,windexaft,with='linespoints',title='recovered')
    #g('set xlabel "alpha"')
    #g('set ylabel "warping index [pixels]"')
    #g.plot(dind,daft)
    g=bigtools.plot(alph,windex,with='linespoints',title='initial',xlabel='alpha',ylabel="warping index [pixels]")
    g("set terminal postscript 20")
    g.hardcopy(datadir+'showstartpinfl-wind.eps')
    g1=bigtools.plot(alph,windexaft,with='linespoints',title='recovered',
                     xlabel='alpha',ylabel="warping index [pixels]")
    g1("set terminal postscript 20")
    g1.hardcopy(datadir+'showstartpinfl-rcvrd.eps')
    raw_input('Press any key')


# ----------------------------------------------------------------------

def try3Dreg():
    """ Let us try to make a simple 3D registration """
    refv=zeros((3,2,5),Float) ; refv[1,1,2]=0.5 ; refv[1,1,3]=0.5
    testv=zeros((3,2,5),Float) ; testv[1,1,3]=1.0 ; 
    par=bigtools.Parameters(None,default={'testimg':testv,'refimg':refv,
                          'degc':1, 'degf':3, 'alternatered':1,
                          'verbose':6, 'crate':(8,8,8),'xtol':1e-10
                          })
    wp=bigfelreg.MultigridableWarpingProblemBase(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    mrs.solveMR()
    x=mrs.getProblem().getXstate() # get the deformation parameters
    c=mrs.getProblem().getX()
    wv=bigfelreg.WarpableImageS(testv,par=par).setDefS(x).getWarped()
    print "refv=",refv
    print "testv=",testv
    print "wv=",array2string(wv,precision=3,suppress_small=1)
    print "c=",c
    
# ----------------------------------------------------------------------

def prepareCT():
    """ Let us prepare the CT tomography images for registration
    We clip the intensities to get rid of the background and also cut down
    to get rid of the dental part and yield two volumes of the same size.
    If changes are made here, the corresponding ones should be made also in
    showCTresults """
    ct0=regultrasound.readRawMatlabInts(
        'data_felreg/philippe-v0-45x512x512-int16.raw',(45,512,512))
    ct1=regultrasound.readRawMatlabInts(
        'data_felreg/philippe-v1-42x512x512-int16.raw',(42,512,512))
    ct0=clip(ct0,0,5000) # clip background
    ct1=clip(ct1,0,5000)
    ct0=reduceCTslices(ct0[0:40])
    ct1r=zeros((40,128,128),Float) 
    ct1r[0:34]=reduceCTslices(ct1[0:34]) ; ct1=ct1r
    f=shelve.open(datadir+"philippe-CT-reduced.dat")
    f['ct0r']=ct0 ; f['ct1r']=ct1
    f.close()

def showCT():
    """ Show some of the CT slices """
    f=shelve.open(datadir+"philippe-CT-reduced.dat")
    ct0=f['ct0r'] ; ct1=f['ct1r']
    f.close()
    v=[]
    for i in [0,5,10,20,30,34,35]:
        v.append(bigtools.view(bigtools.showDif(ct0[i],ct1[i])))
    raw_input('Press any key')
    
        
def reduceCTslices(v):
    """ Reduces the size of the volume CT across x and y dimensions, not z
    Currently four-fold reduction """

    for i in xrange(v.shape[0]):
        print "Reducing slice ", i
        rimg=bigsplines.MSplineSignal(v[i]).reduce().reduce().evalint()
        v[i,0:rimg.shape[0],0:rimg.shape[1]]=rimg
    return v[:,0:rimg.shape[0],0:rimg.shape[1]]

def registerCT():
    """ Register CT volumes """
    f=shelve.open(datadir+"philippe-CT-reduced.dat")
    ct0=f['ct0r'] ; ct1=f['ct1r']
    f.close()
    v0=bigtools.view(bigtools.showDif(ct0[20],ct1[20]),
                     title='difference before',size=vsize)
    par=bigtools.Parameters({'verbose':6, 'crate':(8,8,8),
                             'xtol':0.01, 'degc':2, 'degf':1 ,
                             'alternatered':1, 
                             'testimg':ct0, 'refimg':ct1})
    wp=bigfelreg.MultigridableWarpingProblemBase(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    mrs.solveMR()
    x=mrs.getProblem().getXstate() # get the deformation parameters
    f=shelve.open(datadir+"registerCT-params.dat")
    f['par']=par
    f['x']=x
    f.close()
    ct0w=bigfelreg.WarpableImageS(ct0,par=par).setDefS(x).getWarped()
    v1=bigtools.view(bigtools.showDif(ct0w[20],ct1[20]),
                     title='difference after',size=vsize)
    raw_input('Press any key')


def showCTregresults():
    """ Show the results of the CT registration. Make sure that the
    transformations here are compatible with what we do in prepareCT()
    """
    usereduced=0 ; showbefore=0
    # read images
    if usereduced:
        f=shelve.open(datadir+"philippe-CT-reduced.dat")
        ct0=f['ct0r'] ; ct1=f['ct1r']
        f.close()
    else:
        ct0=regultrasound.readRawMatlabInts(
            'data_felreg/philippe-v0-45x512x512-int16.raw',(45,512,512))
        ct1=regultrasound.readRawMatlabInts(
            'data_felreg/philippe-v1-42x512x512-int16.raw',(42,512,512))
        ct0=clip(ct0,0,5000)[0:40] # clip background
        ct1=clip(ct1,0,5000)[0:40]
        ct1[34:]=zeros(ct1.shape[1:]) # clip bottom part as in prepareCT
    # read parameters
    f=shelve.open(datadir+"registerCT-params.dat")
    par=f['par']
    x=f['x']
    f.close()
    par=bigtools.Parameters(par,override={'degf':1})
    if usereduced: s=1 # scale changes
    else:          s=4 
    # maybe change sizes ? Unfortunately we do not know the voxel size
    ctsize=(512,512)
    if showbefore:
        dv=bigtools.showDif(ct0[20],ct1[20])
        v0=bigtools.view(dv,title='axial view, before',size=ctsize,
                         resample=Image.BILINEAR)
        bigtools.printtofile(dv,datadir+'showctres-axb.tiff')
        dv=bigtools.showDif(ct0[:,s*64],ct1[:,s*64])
        v1=bigtools.view(dv,size=ctsize,title='coronal view, before',
                         resample=Image.BILINEAR)
        bigtools.printtofile(dv,datadir+'showctres-cob.tiff')
        dv=bigtools.showDif(ct0[:,:,s*64],ct1[:,:,s*64])
        v2=bigtools.view(dv,size=ctsize,title='sagital view, before',
                         resample=Image.BILINEAR)
        bigtools.printtofile(dv,datadir+'showctres-sab.tiff')
        raw_input('Press any key')

    if not usereduced:
        par=bigtools.Parameters(par,
            override={'crate':(par.crate*array((1.0,s,s))).astype(Int),
                      'cacheidentity': 0,
                      'verbose': 10})

    print "creating warpable image"
    ct0s=bigfelreg.WarpableImageS(ct0,par=par).setDefS(x)
    print "We have warpable image"
    if not usereduced: # modify deformation parameters
        c=ct0s.getCAsArray()
        c[1:]*=s 
        ct0s.setCFromArray(c)
    ct0w=ct0s.getWarped()
    print "We have warped image"
    dv=bigtools.showDif(ct0w[20],ct1[20])
    va0=bigtools.view(dv,title='axial view, after',size=ctsize,
                     resample=Image.BILINEAR)
    bigtools.printtofile(dv,datadir+'showctres-axa.tiff')
    dv=bigtools.showDif(ct0w[:,s*64],ct1[:,s*64])
    va1=bigtools.view(dv,size=ctsize,title='coronal view, after',
                     resample=Image.BILINEAR)
    bigtools.printtofile(dv,datadir+'showctres-coa.tiff')
    dv=bigtools.showDif(ct0w[:,:,s*64],ct1[:,:,s*64])
    va2=bigtools.view(dv,size=ctsize,title='sagital view, after',
                     resample=Image.BILINEAR)
    bigtools.printtofile(dv,datadir+'showctres-saa.tiff')
    raw_input('Press any key')



# ----------------------------------------------------------------------

def recoverDeformation(img,par,g,mask=None,sigma=0.0,calcwii=0):
    """ Deforms an image img with deformation g. Then try to recover
    it with parameters par. Gaussian noise with stdev sigma
    can be added to images.

    Returns [Eb,wib,Ea,wia,wii,t], where wii is the ideal
    warping index. """
    w=bigfelreg.WarpableImageS(img,par=par)
    imw=w.getWarped(g=g)
    if mask is None:
        npix=bigtools.prodall(img.shape)
        Eb=bigtools.sumall(abs(imw-img)**2)/npix
    else:
        npix=bigtools.sumall(bigtools.int_array(mask))
        Eb=bigtools.sumall(abs((imw-img)*mask)**2)/npix
    wib=w.getWarpingIndexG(g,mask=mask) # warp. ind. with identity  
    print "Eb=",Eb,"  wib=",wib
    if calcwii: # yes, calculate the ideal fit
        w.setProjection(g,mask=mask) # improve fit
        wii=w.getWarpingIndexG(g,mask=mask)
        print "wii=",wii
    else:
        wii=0.0

    del w
    imwn=imw+RandomArray.normal(0.0,sigma,imw.shape)
    print "Starting registration"
    par=bigtools.Parameters(par,override={
            'testimg':img, 'refimg':imwn})
    wp=bigfelreg.MultigridableWarpingProblemBase(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    t=-time.clock() ; mrs.solveMR() ; t+=time.clock()
    print "Registration took %f s." % t
    x=mrs.getProblem().getXstate()
    wp.datapart.testw.setDefS(x)
    wia=wp.datapart.testw.getWarpingIndexG(g,mask=mask)
    imgw=wp.datapart.testw.getWarped()
    if mask is None:
        Ea=bigtools.sumall(abs(imgw-imw)**2)/npix
    else:
        Ea=bigtools.sumall(abs((imgw-imw)*mask)**2)/npix
    print "Eb=",Eb,"  wib=",wib, "wii=", wii, "Ea=",Ea,"  wia=",wia
    return Eb,wib,Ea,wia,wii,t

def tryAllDegrees(img,par,g,mask=None):
    """ makes registration using deformation described by splines of
    degrees 1,2,3

    Returns a list of tuples (degc,Eb,wib,Ea,wia,wii,t)
    """

    resl=[]
    for degc in [1,2,3]:
        par=bigtools.Parameters(par,override={'degc':degc})
        calcwii= (degc==1)
        #calcwii=0 
        Eb,wib,Ea,wia,wiiq,t=recoverDeformation(img,par,g,mask=mask,
                                               sigma=0.0,calcwii=calcwii)
        if calcwii: wii=wiiq
        resl.append((degc,Eb,wib,Ea,wia,wii,t))
    return resl

def genRandDef(J,r,ampl,mask=None):
    """ Generates random 2D deformation with Sobolev exponent r
    and warping index ampl. Size (2**J)**2 """
    M=2**J
    g1=reshape(bigfract.genfractrand2D(r,J),(1,M,M))
    g2=reshape(bigfract.genfractrand2D(r,J),(1,M,M))
    identity=indices((M,M),Float)
    g=concatenate((g1,g2))+identity
    wi=bigfelreg.WarpableImageS(zeros((M,M),Float)).getWarpingIndexG(g,mask=mask)
    g=concatenate((g1,g2))*(ampl/wi)+identity
    wi=bigfelreg.WarpableImageS(zeros((M,M),Float)).getWarpingIndexG(g,mask=mask)
    #print "wi2=",wi
    return g

def showDegcDependence():
    """ Makes experiments to find the good degc """

    resl=[] 
    N=100    # how many runs to make
    sobrlist=[1.0,1.2,1.4,1.6,1.8,2.0,2.2,2.4] # Sobolev exponent list
    ampl=5.0 # mean warping index
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    mask=loadMask()
    par=bigtools.Parameters({'xtol':0.1, 'xtollast': 0.01,
                             'degf':3, 'verbose':4,
                             'crate':(32,32)})
    #par=bigtools.Parameters({'xtol':10.0, 'xtollast': 10.0,
    #                         'degf':3, 'verbose':3,
    #                         'crate':(32,32)})
    for n in xrange(N):
        for r in sobrlist:
            # generate deformation, 8 means image 256x256
            g=genRandDef(8,r,ampl,mask=mask) 
            #grid=bigtools.makeGrid(ima.shape)
            #dg=bigfelreg.WarpableImageS(grid,par=par).getWarped(g)
            # v1=bigtools.view(dg,title='deformed grid, r=%f' % r)
            rl=tryAllDegrees(ima,par,g,mask=mask)
            for item in rl:
                degc,Eb,wib,Ea,wia,wiiq,t=item
                resl.append((n,r,degc,Eb,wib,Ea,wia,wiiq,t))
            f=shelve.open(datadir+"showdegc-resl.dat")
            f['resl']=resl ; 
            f.close()

def showDegcRes():
    """ Make a nice graph out of the showDegcDependence results

        Beware of the hack below. To be corrected.
    """
    f=shelve.open(datadir+"showdegc-resl.dat")
    resl=f['resl'] ; 
    f.close()
    #print resl
    rdegcs={} ; degcs={} ; rs={}
    for x in resl:
        n,r,degc,Eb,wib,Ea,wia,wiiq,t=x
        degcs[degc]=1 ; rs[r]=1
        key=(r,degc)
        if r==1.8 and degc==2:
            print "r=",r, " degc=",degc, "wi=", wia
        values=array((Eb,wib,Ea,wia,wiiq,t),Float)
        if rdegcs.has_key(key): # existing entry
            oldval,cnt=rdegcs[key]
            values+=oldval ; cnt+=1
        else: # new entry
            cnt=1
        rdegcs[key]=(values,cnt)
    # Print according to degree
    degcsl=degcs.keys() ; degcsl.sort() ; print "degcsl=",degcsl
    rsl=rs.keys() ; rsl.sort() ; print "rsl=",rsl
    ndegcsl=len(degcsl) ; nrsl=len(rsl)
    resa=zeros((ndegcsl,nrsl,6),Float)
    for di in xrange(ndegcsl):
        for ri in xrange(nrsl):
            degc=degcsl[di] ; r=rsl[ri]
            val,cnt=rdegcs[(r,degc)]
            resa[di,ri]=val/cnt
            #print "degc=",degc," r=",r, " cnt=",cnt

            
    # make the plots
    g=Gnuplot.Gnuplot() # first the SSD per pixel
    #dEb=Gnuplot.Data(rsl,resa[0,:,0],with='lines',title='before')
    dEa1=Gnuplot.Data(rsl,resa[0,:,2],with='linespoints',title='degree 1')
    dEa2=Gnuplot.Data(rsl,resa[1,:,2],with='linespoints',title='degree 2')
    dEa3=Gnuplot.Data(rsl,resa[2,:,2],with='linespoints',title='degree 3')
    g.xlabel('r') ; g.ylabel('SSD per pixel')
    g("set key left bottom")
    g.plot(dEa1,dEa2,dEa3)
    g("set terminal postscript 20")
    g.hardcopy(datadir+"showdegcres-E.eps")

    print "deg3", resa[2,:,3]
    print "ideal", resa[0,:,4]
    # HERE IS THE HACK to avoid recalculating everything
    # the correct solution is to reimplement setProjection to be noniterative
    resa[0,5:,4]=array((0.45,0.28,0.22))
    print "ideal", resa[0,:,4]

    g1=Gnuplot.Gnuplot() # Then the wi
    #dwb=Gnuplot.Data(rsl,resa[0,:,1],with='lines',title='before')
    dwi=Gnuplot.Data(rsl,resa[0,:,4],with='linespoints',title='ideal')
    dwa1=Gnuplot.Data(rsl,resa[0,:,3],with='linespoints',title='degree 1')
    dwa2=Gnuplot.Data(rsl,resa[1,:,3],with='linespoints',title='degree 2')
    dwa3=Gnuplot.Data(rsl,resa[2,:,3],with='linespoints',title='degree 3')
    g1.xlabel('r') ; g1.ylabel('warping index [pixels]')
    g1("set key left bottom")
    # g1("set yrange [0:6.0]")
    g1.plot(dwi,dwa1,dwa2,dwa3)
    g1("set terminal postscript 20")
    g1.hardcopy(datadir+"showdegcres-wi.eps")

    g2=Gnuplot.Gnuplot() # Then the time
    dt1=Gnuplot.Data(rsl,resa[0,:,5],with='linespoints',title='degree 1')
    dt2=Gnuplot.Data(rsl,resa[1,:,5],with='linespoints',title='degree 2')
    dt3=Gnuplot.Data(rsl,resa[2,:,5],with='linespoints',title='degree 3')
    g2.xlabel('r') ; g2.ylabel('time [s]')
    #g1("set key left bottom")
    g2("set yrange [0:200]")
    g2.plot(dt1,dt2,dt3)
    g2("set terminal postscript 20")
    g2.hardcopy(datadir+"showdegcres-time.eps")
    raw_input('Press any key')
            
# ----------------------------------------------------------------------

def showCrateDep():
    """ Makes experiments to investigate the effects of crate """
    resl=[] 
    N=100    # how many runs to make
    sobrlist=[1.0,1.2,1.4,1.6,1.8,2.0,2.2,2.4] # Sobolev exponent list
    ampl=5.0 # mean warping index
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    mask=loadMask()
    par=bigtools.Parameters({'xtol':0.1, 'xtollast': 0.01,
                             'degf':3, 'degc':3, 'verbose':4})
    for n in xrange(N):
        for r in sobrlist:
            print "showCrateDep n=",n," r=",r
            # generate deformation, 8 means image 256x256
            g=genRandDef(8,r,ampl,mask=mask) 
            #grid=bigtools.makeGrid(ima.shape)
            #dg=bigfelreg.WarpableImageS(grid,par=par).getWarped(g)
            # v1=bigtools.view(dg,title='deformed grid, r=%f' % r)
            rl=tryAllCrates(ima,par,g,mask=mask)
            for item in rl:
                crate,Eb,wib,Ea,wia,wiiq,t=item
                resl.append((n,r,crate,Eb,wib,Ea,wia,wiiq,t))
            f=shelve.open(datadir+"showcrate-resl.dat")
            f['resl']=resl ; 
            f.close()

def tryAllCrates(img,par,g,mask=None):
    """ makes registration using deformation described by splines
    with different crates

    Returns a list of tuples (degc,Eb,wib,Ea,wia,wii,t)
    """

    resl=[]
    for crate in [256,128,64,32,16,8]:
        par=bigtools.Parameters(par,override={'crate':(crate,crate)})
        calcwii= (crate==256) # the first item
        #calcwii=0 
        Eb,wib,Ea,wia,wiiq,t=recoverDeformation(img,par,g,mask=mask,
                                               sigma=0.0,calcwii=calcwii)
        if calcwii: wii=wiiq
        resl.append((crate,Eb,wib,Ea,wia,wii,t))
    return resl

def showCrateRes():
    """ Make a nice graph out of the showCrate results """
    f=shelve.open(datadir+"showcrate-resl.dat")
    resl=f['resl'] ; 
    f.close()
    #print resl
    rcrates={} ; crates={} ; rs={}
    for x in resl:
        n,r,crate,Eb,wib,Ea,wia,wiiq,t=x
        crates[crate]=1 ; rs[r]=1
        key=(r,crate)
        #print "r=",r, " crate=",crate
        values=array((Eb,wib,Ea,wia,wiiq,t),Float)
        if rcrates.has_key(key): # existing entry
            oldval,cnt=rcrates[key]
            values+=oldval ; cnt+=1
        else: # new entry
            cnt=1
        rcrates[key]=(values,cnt)
    # Sort according to crate
    cratesl=crates.keys() ; cratesl.sort() ; print "cratesl=",cratesl
    rsl=rs.keys() ; rsl.sort() ; print "rsl=",rsl
    ncratesl=len(cratesl) ; nrsl=len(rsl)
    resa=zeros((ncratesl,nrsl,6),Float)
    for di in xrange(ncratesl):
        for ri in xrange(nrsl):
            crate=cratesl[di] ; r=rsl[ri]
            val,cnt=rcrates[(r,crate)]
            resa[di,ri]=val/cnt
            
    # make the plots
    g=Gnuplot.Gnuplot() # first the SSD per pixel
    dE=[Gnuplot.Data(rsl,resa[0,:,0],with='lines',title='before')]
    for i in xrange(ncratesl):
        dE.append(Gnuplot.Data(rsl,resa[i,:,2],with='linespoints',
                               title='crate %d' % cratesl[i]))
    g.xlabel('r') ; g.ylabel('SSD per pixel')
    g("set key left bottom")
    apply(g.plot,dE)
    g("set terminal postscript 20")
    g.hardcopy(datadir+"showcrateres-E.eps")
    

    g1=Gnuplot.Gnuplot() # Then the wi
    dw=[Gnuplot.Data(rsl,resa[0,:,1],with='lines',title='before')]
    for i in xrange(ncratesl):
        dw.append(Gnuplot.Data(rsl,resa[i,:,3],with='linespoints',
                               title='crate %d' % cratesl[i]))
    g1.xlabel('r') ; g1.ylabel('warping index [pixels]')
    g1("set key left bottom")
    g1("set yrange [0:6.0]")
    apply(g1.plot,dw)
    g1("set terminal postscript 20")
    g1.hardcopy(datadir+"showcrateres-wi.eps")
    raw_input('Press any key')

# ----------------------------------------------------------------------

def showNoiseDep():
    """ Makes experiments to investigate the effects of noise """
    resl=[] 
    N=100    # how many runs to make
    r=2.0
    ampl=5.0 # mean warping index
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    mask=loadMask()
    par=bigtools.Parameters({'xtol':0.1, 'xtollast': 0.01,
                             'degf':3, 'degc':3, 'verbose':4,
                             'crate':(32,32)})
    for n in xrange(N):
            print "showNoiseDep n=",n,
            g=genRandDef(8,r,ampl,mask=mask) 
            rl=tryAllNoises(ima,par,g,mask=mask)
            for item in rl:
                sigmadb,Eb,wib,Ea,wia,wiiq,t=item
                resl.append((n,sigmadb,Eb,wib,Ea,wia,wiiq,t))
            f=shelve.open(datadir+"shownoisedep-resl.dat")
            f['resl']=resl ; 
            f.close()

def tryAllNoises(img,par,g,mask=None):
    """ makes registration using deformation described by splines
    with different noise levels

    Returns a list of tuples (sigma,Eb,wib,Ea,wia,wii,t)
    """

    resl=[] ; sigmaDBlist=[ -30, -20, -10, 0]
    for i in xrange(len(sigmaDBlist)):
        sigmaDB=sigmaDBlist[i]
        sigma=255.0*10.0**(sigmaDB/20.0)
        print "tryAllNoises sigmaDB=",sigmaDB
        #calcwii= (i==0) # the first item
        calcwii=0 ; wii=0.0
        Eb,wib,Ea,wia,wiiq,t=recoverDeformation(img,par,g,mask=mask,
                                               sigma=sigma,calcwii=calcwii)
        if calcwii: wii=wiiq
        resl.append((sigmaDB,Eb,wib,Ea,wia,wii,t))
    return resl

def showNoisesRes():
    """ Make a nice graph out of the showNoisesDep results """
    f=shelve.open(datadir+"shownoisedep-resl.dat")
    resl=f['resl'] ; 
    f.close()
    #print resl
    sdbs={} ; rnoise={}
    for x in resl:
        n,sdb,Eb,wib,Ea,wia,wiiq,t=x
        sdbs[sdb]=1 ; 
        key=sdb
        #print "r=",r, " crate=",crate
        values=array((Eb,wib,Ea,wia,wiiq,t),Float)
        if rnoise.has_key(key): # existing entry
            oldval,cnt=rnoise[key]
            values+=oldval ; cnt+=1
        else: # new entry
            cnt=1
        rnoise[key]=(values,cnt)
    # Print according to degree
    sdbsl=sdbs.keys() ; sdbsl.sort() ; print "sdbsl=",sdbsl
    nsdbsl=len(sdbsl) 
    resa=zeros((nsdbsl,6),Float)
    for sdbi in xrange(nsdbsl):
            sdb=sdbsl[sdbi] 
            val,cnt=rnoise[sdb]
            resa[sdbi]=val/cnt

    stdev=zeros((nsdbsl,6),Float)
    for sdbi in xrange(nsdbsl):
        for x in resl:
            n,sdb,Eb,wib,Ea,wia,wiiq,t=x
            if sdb!=sdbsl[sdbi]: continue
            values=array((Eb,wib,Ea,wia,wiiq,t),Float)
            stdev[sdbi]+=(values-resa[sdbi])**2
        stdev[sdbi]/=rnoise[sdb][1]
        stdev=sqrt(stdev)

    print "stdev E", stdev[:,2], " stdev wi", stdev[:,3]
    
    # make the plots
    g=Gnuplot.Gnuplot() # first the SSD per pixel
    dE=Gnuplot.Data(sdbsl,resa[:,2],with='lines')
    dEeb=Gnuplot.Data(sdbsl,resa[:,2],stdev[:,2],with='errorbars')
    g.xlabel('SNR [dB]') ; g.ylabel('SSD per pixel')
    g("set xrange [-35:5]" )
    g.plot(dE,dEeb)
    g("set terminal postscript 20")
    g.hardcopy(datadir+"shownoiseres-E.eps")

    g1=Gnuplot.Gnuplot() # then the wi
    dE=Gnuplot.Data(sdbsl,resa[:,3],with='linespoints')
    dEeb=Gnuplot.Data(sdbsl,resa[:,3],stdev[:,3],with='errorbars')
    g1.xlabel('SNR [dB]') ; g1.ylabel('warping index [pixels]')
    g1("set xrange [-35:5]" )
    g1.plot(dE,dEeb)
    g1("set terminal postscript 20")
    g1.hardcopy(datadir+"shownoiseres-wi.eps")
    raw_input('Press any key')

# ----------------------------------------------------------------------


def showFractDef():
    """ Demonstrates fractal deformations """
    sobrlist=[1.0,1.2,1.4,1.6,1.8,2.0,2.2,2.4] # Sobolev exponent list
    ampl=5.0 # mean warping index
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    mask=loadMask()
    par=bigtools.Parameters({'degf':3})

    for r in sobrlist:
        # generate deformation, 8 means image 256x256
        g=genRandDef(8,r,ampl,mask=mask) 
        grid=bigtools.makeGrid(ima.shape)
        dg=bigfelreg.WarpableImageS(grid,par=par).getWarped(g)
        #v1=bigtools.view(dg,title='deformed grid, r=%0.2f' % r)
        #raw_input('Press any key')
        bigtools.printtofile(dg,datadir+"showfractdef%0.2f.tiff" % r)
        
# ----------------------------------------------------------------------

def test_scalechange():
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    parsmall=bigtools.Parameters({'crate':(8,8)})
    s=4 # scale factor
    parbig=bigtools.Parameters({'crate':
                          (parsmall.crate*array((s,s))).astype(Int)})
    wi=bigsplines.MSplineSignal(ima)
    imasmall=wi.reduce().reduce().evalint()
    #print imasmall.shape
    wismall=bigfelreg.WarpableImageS(imasmall,par=parsmall)
    wibig=bigfelreg.WarpableImageS(ima,par=parbig)
    #imase=wismall.expand().expand().evalint()
    #print bigtools.mdownsample(ima,(4,4)).shape
    #df=bigtools.showDif(imasmall,bigtools.mdownsample(ima,(4,4)))
    #df=imasmall-bigtools.mdownsample(ima,(4,4))
    #v1=bigtools.view(df,size=(512,512))
    #raw_input('Press any key')
    c=wismall.getCAsArray() ; print "c.shape=",c.shape
    c[0,1:4,3:5]=10.0 ; c[1,5:8,5:7]=-15.0 
    c[1,:,:]+=4.0 # shift down 
    wismall.setCFromArray(c)
    imwsmall=wismall.getWarped()
    c*=s
    c1=wibig.getCAsArray() ; print "c1.shape=",c1.shape
    wibig.setCFromArray(c)
    imwbig=wibig.getWarped()
    df=bigtools.showDif(imwsmall,imasmall)
    v1=bigtools.view(df,size=(512,512))
    df=bigtools.showDif(imwbig,ima)
    v2=bigtools.view(df,size=(512,512))
    raw_input('Press any key')

# ----------------------------------------------------------------------

def register_anat():
    imr=bigtools.ImageToArray(Image.open(datadir+'anatMRI-jk-83.tiff'))
    imr=transpose(imr)
    #print "imr ", bigtools.maxelement(imr) 
    imt=bigtools.ImageToArray(Image.open(datadir+'anatMRI-nirkko-18.tiff'))
    #print "imt ", bigtools.maxelement(imt)
    vsize=(512,512)
    v1=bigtools.view(imr,title='reference',size=vsize)
    v2=bigtools.view(imt,title='test',size=vsize)
    v3=bigtools.view(bigtools.showDif(imr,imt),title='difference before',
                     size=(512,512))
    bigtools.printtofile(bigtools.showDif(imr,imt),
                         datadir+'reganat-difb.tiff')
    par=bigtools.Parameters({'verbose':6, 'crate':(64,64),
                             'xtol':0.1, 'xtollast': 0.01, 'degc':2, 'degf':1 ,
                             'alternatered':1, 'grmask':0,
                             'testimg':imt, 'refimg':imr})
    wp=bigfelreg.MultigridableWarpingProblemBase(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    mrs.solveMR()
    x=mrs.getProblem().getXstate() # get the deformation parameters
    imtw=bigfelreg.WarpableImageS(imt,par=par).setDefS(x).getWarped()
    v4=bigtools.view(imtw,title='Warped test image',size=vsize)
    #bigtools.printtofile(img2w,datadir+'reganat-imtw.tiff')
    v5=bigtools.view(bigtools.showDif(imr,imtw),
                     title='difference after',size=vsize)
    bigtools.printtofile(bigtools.showDif(imr,imtw),
                         datadir+'reganat-difa.tiff')
    grid=bigtools.makeGrid(imr.shape)
    dg=bigfelreg.WarpableImageS(grid,par=par).setDefS(x).getWarped()
    v6=bigtools.view(dg,title='deformed grid',size=vsize)
    bigtools.printtofile(dg,datadir+'reganat-gridw.tiff')
    raw_input('Press any key')


# ----------------------------------------------------------------------

def register_mriheart():
    """ Try to register MRI heart images """
    doregister=0
    imr=bigtools.ImageToArray(Image.open(datadir+'heart-mri-1cut.tiff'))[0]
    #imr=transpose(imr)
    #print "imr ", imr.shape
    imt=bigtools.ImageToArray(Image.open(datadir+'heart-mri-5cut.tiff'))[0]
    #print "imt ", imt.shape
    vsize=(512,512)
    #v1=bigtools.view(imr,title='reference',size=vsize)
    #v2=bigtools.view(imt,title='test',size=vsize)
    rcont=bigtools.ImageToArray(Image.open(
        datadir+'heart-mri-1cut-edges.png'))>100
    #v2a=bigtools.view(rcont,title='contours',size=vsize)
    difimg=bigtools.showContours(imr,rcont)
    v2b=bigtools.view(difimg,size=vsize,title='contours+reference')
    bigtools.printtofile(difimg,datadir+'regmriheart-cimr.tiff')
    difimg=bigtools.showContours(imt,rcont)
    v2c=bigtools.view(difimg,size=vsize,title='contours+test')
    bigtools.printtofile(difimg,datadir+'regmriheart-cimt.tiff')

    #v3=bigtools.view(bigtools.showDif(imr,imt),title='difference before',
    #                 size=(512,512))
    raw_input('Press any key')

    #bigtools.printtofile(bigtools.showDif(imr,imt),
    #                     datadir+'regmriheart-difb.tiff')
    par=bigtools.Parameters({'verbose':6, 'crate':(32,32),
                             'xtol':0.1, 'xtollast': 0.01, 'degc':2, 'degf':1 ,
                             'alternatered':1, 'grmask':0,
                             'testimg':imt, 'refimg':imr})
    if doregister:
        wp=bigfelreg.MultigridableWarpingProblemBase(par=par)
        mrs=bigoptimize.MROptimizerState(wp,par=par)
        mrs.solveMR()
        x=mrs.getProblem().getXstate() # get the deformation parameters
        f=shelve.open(datadir+"mriheart-results.dat")
        f['x']=x
        f.close()
    else:
        f=shelve.open(datadir+"mriheart-results.dat")
        x=f['x']
        f.close()
        
    imtw=bigfelreg.WarpableImageS(imt,par=par).setDefS(x).getWarped()
    v4=bigtools.view(imtw,title='Warped test image',size=vsize)
    #bigtools.printtofile(img2w,datadir+'reganat-imtw.tiff')
    #v5=bigtools.view(bigtools.showDif(imr,imtw),
    #                 title='difference after',size=vsize)
    #bigtools.printtofile(bigtools.showDif(imr,imtw),
    #                     datadir+'regmriheart-difa.tiff')
    difimg=bigtools.showContours(imtw,rcont)
    v2d=bigtools.view(difimg,size=vsize,title='contours+warped test')
    bigtools.printtofile(difimg,datadir+'regmriheart-cimtw.tiff')

    grid=bigtools.makeGrid(imr.shape)
    dg=bigfelreg.WarpableImageS(grid,par=par).setDefS(x).getWarped()
    v6=bigtools.view(dg,title='deformed grid',size=vsize)
    bigtools.printtofile(dg,datadir+'regmriheart-gridw.tiff')
    raw_input('Press any key')


# ----------------------------------------------------------------------

def register_harv():
    """ Register VisibleHuman proton density slice with Harvard medical
    atlas image """
    imr=bigtools.ImageToArray(Image.open(datadir+'harvbrain-pd-023.tiff'))
    #imr=transpose(imr)
    print "imr ", bigtools.maxelement(imr), imr.shape
    # histogram equalized version
    imt=bigtools.ImageToArray(Image.open(datadir+'vishum-head-mri-unl-ext-heq.tiff'))
    print "imt ", bigtools.maxelement(imt), imt.shape
    vsize=(512,512)
    v1=bigtools.view(imr,title='reference')
    v2=bigtools.view(imt,title='test')
    #raw_input('Press any key')

    
    v3=bigtools.view(bigtools.showDif(imr,imt),title='difference before',
                     size=(512,512))
    bigtools.printtofile(bigtools.showDif(imr,imt),
                         datadir+'regharv-difb.tiff')
    par=bigtools.Parameters({'verbose':6, 'crate':(32,32),
                             'xtol':0.5, 'xtollast': 0.1, 'degc':3, 'degf':3 ,
                             'alternatered':1, 'grmask':0,
                             'testimg':imt, 'refimg':imr})
    wp=bigfelreg.MultigridableWarpingProblemBase(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    mrs.solveMR()
    x=mrs.getProblem().getXstate() # get the deformation parameters
    imtw=bigfelreg.WarpableImageS(imt,par=par).setDefS(x).getWarped()
    v4=bigtools.view(imtw,title='Warped test image',size=vsize)
    #bigtools.printtofile(img2w,datadir+'reganat-imtw.tiff')
    v5=bigtools.view(bigtools.showDif(imr,imtw),
                     title='difference after',size=vsize)
    bigtools.printtofile(bigtools.showDif(imr,imtw),
                         datadir+'regharv-difa.tiff')
    grid=bigtools.makeGrid(imr.shape)
    dg=bigfelreg.WarpableImageS(grid,par=par).setDefS(x).getWarped()
    v6=bigtools.view(dg,title='deformed grid',size=vsize)
    bigtools.printtofile(dg,datadir+'regharv-gridw.tiff')
    raw_input('Press any key')


# ----------------------------------------------------------------------
def register_harv_landmarks():
    """ Register VisibleHuman proton density slice with Harvard medical
    atlas image using landmarks"""
    imr=bigtools.ImageToArray(Image.open(datadir+'harvbrain-pd-023.tiff'))
    #imr=transpose(imr)
    print "imr ", bigtools.maxelement(imr), imr.shape
    # histogram equalized version
    imt=bigtools.ImageToArray(Image.open(datadir+'vishum-head-mri-unl-ext-heq.tiff'))
    print "imt ", bigtools.maxelement(imt), imt.shape
    vsize=(512,512)
    # give landmarks in the original coordinates, (y,x) ordering
    # landmark(s) in the reference image
    lnd1=array(((464,249),(292,80),(298,403),(111,174),(111,308),
                (179,164),(175,323),(407,125)),Float) 
    # landmark(s) in image 2
    lnd2=array(((451,253),(306,119),(299,394),(144,187),(140,311),
                (201,188),(196,319),(404,156)),Float)
    lndf=array((1.0,1.0,1.0,1.0,1.0,1.0,1.0,0.2),Float)
    v1=bigtools.view(bigtools.showImgWithLandmarks(imr,lnd1,lndsize=4),
                     title='Reference',size=vsize)
    bigtools.printtofile(bigtools.showImgWithLandmarks(imr,lnd1,lndsize=4),
                         datadir+'regharv-lndref.tiff')
    v2=bigtools.view(bigtools.showImgWithLandmarks(imt,lnd2,lndsize=4),
                     title='Test',size=vsize)
    bigtools.printtofile(bigtools.showImgWithLandmarks(imt,lnd2,lndsize=4),
                         datadir+'regharv-lndtest.tiff')
    #raw_input('Press any key')

    par=bigtools.Parameters({'verbose':6, 'crate':(64,64),
                             'xtol':0.1, 'xtollast': 0.1, 'degc':3, 'degf':3 ,
                             'alternatered':1, 'grmask':0,
                             'springpos': lnd1, 'springtarget':lnd2,
                             'springfactor':lndf, 'minlevel':0, 'maxlevel':6,
                             'testimg':imt, 'refimg':imr})
    wp=bigfelreg.MultigridableWarpingProblemSprings(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    mrs.solveMR()
    x=mrs.getProblem().getXstate() # get the deformation parameters
    imtw=bigfelreg.WarpableImageS(imt,par=par).setDefS(x).getWarped()
    v4=bigtools.view(imtw,title='Warped test image',size=vsize)
    #bigtools.printtofile(img2w,datadir+'reganat-imtw.tiff')
    v5=bigtools.view(bigtools.showDif(imr,imtw),
                     title='difference after',size=vsize)
    bigtools.printtofile(bigtools.showDif(imr,imtw),
                         datadir+'regharv-lnddifa.tiff')
    grid=bigtools.makeGrid(imr.shape)
    dg=bigfelreg.WarpableImageS(grid,par=par).setDefS(x).getWarped()
    v6=bigtools.view(dg,title='deformed grid',size=vsize)
    bigtools.printtofile(dg,datadir+'regharv-lndgridw.tiff')
    raw_input('Press any key')


# ----------------------------------------------------------------------

def makeSequence():
    """ This experiment makes a sequence of images (a little film)
    showing the evolution of the algorithm over time """
    ima=bigtools.ImageToArray(Image.open(datadir+'AnatMRI-jk-016.tiff'))
    asize=2*array(ima.shape)
    # verbosity may not be higher than 4 if we want nice graphs !
    # orig settings (used in the article were 0.5 and 0.1 (xtollast)
    par=bigtools.Parameters({'xtol':0.1, 'xtollast': 0.1,
                             'verbose':7, 'crate':(32,32),
                             'maxlevel':5})
    wi=bigfelreg.WarpableImageS(ima,par=par)
    c=wi.getCAsArray() ; print "c.shape=",c.shape
    c[0,1:4,3:5]=10.0 ; c[1,5:8,5:7]=-15.0 
    c[1,:,:]+=4.0 # shift down 
    wi.setCFromArray(c)
    imw=wi.getWarped()
    mask=loadMask()
    par=bigtools.Parameters(par,default={'testimg':ima,'refimg':imw,
                          'degc':2, 'degf':1, 'alternatered':1,      
                          'fileseqmask': datadir+'makeseq%04d.tiff',
                          'imgsize':(256,256)})
    wp=bigfelreg.MultigridableWarpingProblemSQ(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    bigtools.timeop(lambda mrs=mrs: mrs.solveMR(),'solveMR')
    x=mrs.getProblem().getXstate()
    cf=mrs.getOptimizer().getX()
    #print cf
    print "cdif=", bigtools.maxelement(abs(c-cf))
    #wp.datapart.testw.setCFromArray(cf)
    print "wi=",wp.datapart.testw.getWarpingIndexDefS(defs=wi.getDefS(),
                                                      par=par,mask=mask)

# ----------------------------------------------------------------------

if __name__=='__main__': # running interactively
    registerDistant()
    registerDistantLandmarks()
