""" Module assembles code to work with and register
    the ultrasound images of MJL

    Jan Kybic, January 2001
    Maria J. Ledesma-Carbayo, January 2001
    $Id: regultrasound.py,v 1.2 2002/01/21 16:33:19 jkybic Exp $
    """

from Numeric import *
import Image
import shelve
import Gnuplot
import bigtools
import bigpyramid
import bigregister
import bigmultigrid

# ------------------------------------------------------------------------
imdir="images_chus/Imn-Chus-4C-Clip0009/"   # where are the images
#imdir="images_chus/Imh-2C-Clip00023-384x288/"   # where are the images
#imdir="images_chus/Imh-2C-Clip0023-384x288/"   # where are the images
#imdir="images_chus/Im-2c-Clip0013-masked/results-grid16x16/"   # where are the images
#imdirjan="images_chus/Im-2c-Clip0013-masked/"
#imnamejan1=imdirjan+"2C-Clip0013-768x576-14_mask.raw" 
#imnamejan2=imdirjan+"2C-Clip0013-768x576-18_mask.raw"
imname1=imdir+"2C-Clip0013-768x576-14_mask.raw" # example two images
imname2=imdir+"2C-Clip0013-768x576-18_mask.raw"
rootfilename=imdir+"Clip0009-768x576-"  # example of sequence root name
nimages=24 # example of sequence root namenumbeer of images(including the first one at the end(21+1))
grname="images_chus/otherimages/grid_768x576.tif"
imsize=(576,768) # typical size of the images, (height,width)
#imsize=(288,384)
scalefactor=0.5 # to generate smaller images
imsize_tiff=(int(576*scalefactor),int(768*scalefactor))
if scalefactor==0.5: #atencion esto es porque la imagen es de este tamanyo
    grname="images_chus/otherimages/grid_384x288.tif"

# ------------------------------------------------------------------------
def readRawMatlabBytes(filename,size):
    """ Return a 2D matrix of size 'size' (height,width) read
        from 'filename'. The returned matrix is unsigned bytes
        Each (unsigned) byte represents one element. 
    """

    f=open(filename,'rb')
    m=fromstring(f.read(),UnsignedInt8)
    m=asarray(reshape(m,size),Float) # upcast, maybe wasteful but easier
    return m[::-1] # turn upside down

# ------------------------------------------------------------------------

def readRawMatlabInts(filename,size):
    """ Return a 1-3D matrix of size 'size', read from 'filename'.
    File contains 16 bit signed integers.
    """
    f=open(filename,'rb')
    m=fromstring(f.read(),Int16).byteswapped()
    m=asarray(reshape(m,size),Float) # upcast, maybe wastful but easier
    return m # turn upside down

# ------------------------------------------------------------------------

def test_readMatlabRawBytes():
    m=readRawMatlabBytes(im2,imsize)
    bigtools.view(m)
    
# ------------------------------------------------------------------------

def register_pair(img1,img2, cshape=(2,6,6), tol=1e-3):
    """ Takes two images (matrices) and registers them.

        Returns the coefficients c, warped img2, and the
        MultigridableWarpingProblem object
        img1 is the reference, img2 test (what we warp)
    """

    degf=3 ; degc=2 # B-spline degrees
    refp=bigpyramid.BSPyramid(img1,degree=degf) # calculate pyramids
    testp=bigpyramid.BSPyramid(img2,degree=degf)
    #cshape=(2,16,16) # the size of the control grid
    #m=bigregister.calculateMask(img1,20) # 20% of pixels to retain
    #bigtools.view(m)
    
    m=None
    # create the warping problem
    mwp=bigregister.MultigridableWarpingProblem(refp,testp,cshape,degf,degc,
                                               mask=m)
    # create the optimizer
    mopt=bigmultigrid.MultigridOptimizerState(mwp,verbose=5,xtol=tol)
    # optimize, measuring the time
    # cnorm=mopt.solveMR()
    cnorm=bigtools.timeop(lambda opt=mopt: opt.solveMR(),'Registration')
    print "Iterations in solveMR ",mopt.itercounts
    c=mwp.denormalizeC(cnorm) # go back to pixel coordinates
    mwp.w.w.setC(c) # warp the test image according to the results
    mwp.w.w.mask=None # full warping now
    img2w=mwp.w.w.getWarped() # warped test image
    return c,img2w,mwp

# ------------------------------------------------------------------------

def register_pair_and_show(img1,img2, cshape=(2,6,6), tol=1e-3):
    """ Takes two images (matrices), registers them and shows the results.
        img1 is the reference, img2 test (what we warp)
    """

    #bigtools.view(img1,title='Reference image')
    #bigtools.view(img2,title='Test image')
    c,img2w,mwp=register_pair(img1,img2, cshape, tol)
    #img2w=img2 ; c=zeros((2,8,8))
    bigtools.view(img2w,title='Warped test image')    
    bigtools.printtofile(img2w,'warped18.tiff')
    wd=abs(img1-img2) ; mwd=bigtools.maxelement(wd) ;
    wda=abs(img1-img2w) ; mwda=bigtools.maxelement(wda) ;
    f=float(max(mwd,mwda)) ; 
    bigtools.view(wd,title='Difference before',scale=(0,f))
    bigtools.view(wda,title='Difference after',scale=(0,f))
    bigtools.printtofile(wd,'diffbefore14-18.tiff',scale=(0,f))
    bigtools.printtofile(wda,'diffafter14-18.tiff',scale=(0,f))
    imgrid=Image.open(grname)
    matgrid=bigtools.ImageToArray(imgrid) # convert to matrix
    wi=bigregister.WarpableImage(matgrid,degc=mwp.degc)
    wi.setC(c)
    defgrid=wi.getWarped()
    bigtools.view(defgrid,title='Deformed grid')
    bigtools.printtofile(defgrid,'defgrid14-18.tiff')

# ------------------------------------------------------------------------

def register_pair_and_save(img1,img2,foutroot, cshape=(2,6,6), tol=1e-3): # CHUS ??? Introducir contornos/  nombre de los contornos
    """ Takes two images (matrices), registers them and saves the results.
        img1 is the reference, img2 test (what we warp)
    """

    #bigtools.view(img1,title='Reference image')
    #bigtools.view(img2,title='Test image')
    c,img2w,mwp=register_pair(img1,img2,cshape, tol)
    #img2w=img2 ; c=zeros((2,8,8))
    #bigtools.view(img2w,title='Warped test image')    
    bigtools.printtofile(img2w,foutroot+"-w.tiff")
    img1=img1.astype(Float)
    wd=img1-img2 ; mwd=bigtools.maxelement(wd) ;
    wda=img1-img2w ; mwda=bigtools.maxelement(wda) ;
    f=float(max(mwd,mwda)) ; 
    #bigtools.view(wd,title='Difference before',scale=(0,f))
    #bigtools.view(wda,title='Difference after',scale=(0,f))
    bigtools.printtofile(wd,foutroot+"-dif-b.tiff")
    bigtools.printtofile(wda,foutroot+"-dif-a.tiff")
    imgrid=Image.open(grname)
    matgrid=bigtools.ImageToArray(imgrid) # convert to matrix
    wi=bigregister.WarpableImage(matgrid,degc=mwp.degc)
    wi.setC(c)
    defgrid=wi.getWarped()
    #bigtools.view(defgrid,title='Deformed grid')
    bigtools.printtofile(defgrid,foutroot+"-grid.tiff")
    #We also save the coefficients
    f=shelve.open(foutroot+"-coeff.dat")
    f['coeff']=c
    f.close()

    #Varias cosas utiles
    #tbl=wi.getDeformation()
    #CHUS????para los contornos
    #imcolor=zeros((3,imsize[0],imsize[1]))
    #imcolor[0]=img1

    
# ------------------------------------------------------------------------

def register_sequence(rootf,nim,cshape=(2,6,6), tol=1e-3, reverse=0):
    """ Takes all the images in the sequence and registers them by pairs.
        img1 is the reference, img2 test (what we warp)
    """
    for i in range(nim-1):
      if reverse:
        imname_test=rootf+`(i+1)`+"_s.tiff"
        imname_ref=rootf+`i`+"_s.tiff" 
      else:
        imname_test=rootf+`i`+"_s.tiff"
        imname_ref=rootf+`(i+1)`+"_s.tiff"
        
      #img1=readRawMatlabBytes(imname_ref,imsize)
      #img2=readRawMatlabBytes(imname_test,imsize)
      im1=Image.open(imname_ref)
      im2=Image.open(imname_test)
      img1=bigtools.ImageToArray(im1)
      img2=bigtools.ImageToArray(im2)
      if reverse:
        foutroot=rootf+`(i)`+"-"+`(i+1)`
      else:
        foutroot=rootf+`(i+1)`+"-"+`(i)`

      register_pair_and_save(img1,img2, foutroot,cshape, tol)

# ------------------------------------------------------------------------
def register_sequence_with_first(rootf,nim,cshape=(2,6,6), tol=1e-3):
    """ Takes all the images in the sequence and registers them by pairs.
        img1 is the reference, img2 test (what we warp)
    """
    for i in range(nim-1): 
        imname_test=rootf+`0`+"_s.tiff"
        imname_ref=rootf+`(i+1)`+"_s.tiff"
        #img1=readRawMatlabBytes(imname_ref,imsize)
        #img2=readRawMatlabBytes(imname_test,imsize)
        im1=Image.open(imname_ref)
        im2=Image.open(imname_test)
        img1=bigtools.ImageToArray(im1)
        img2=bigtools.ImageToArray(im2)
        foutroot=rootf+`(i+1)`+"-"+`(0)`
        register_pair_and_save(img1,img2, foutroot, cshape, tol)

# ------------------------------------------------------------------------
def resizing_sequence(rootf,nim,scalefactor):
    """ Takes all the images in the sequence and registers them by pairs.
        img1 is the reference, img2 test (what we warp)
    """
    for i in range(nim): 
        imname_in=rootf+`i`+"_mask.raw"
        img_in=readRawMatlabBytes(imname_in,imsize)
        im=bigtools.ArrayToImage(img_in)
        im_out=im.resize((int(scalefactor*imsize[1]),int(scalefactor*imsize[0])))
        fileout=rootf+`i`+"_s.tiff"
        im_out.save(fileout)
        #bigtools.printtofile(im_out.,fileout)
  
# ------------------------------------------------------------------------

def deformation_sequence(rootf,nim,point,reverse=0):
    """ Takes all the images all the deformations from the deforamtion tables
    and plots the cumulative deformation.
    """
   
    pointdef=zeros((nim,2),Float)
    #we calculate the cumulative deformation
    pointdefcum=zeros((nim,2),Float) 
    img=zeros(imsize_tiff)
    wim=bigregister.WarpableImage(img,imageiscoefs=1)
    for i in range(nim-1):
        if reverse:
           coeff_file=rootf+`(nim-2-i)`+"-"+`(nim-1-i)`+'-coeff.dat' 
        else:
           coeff_file=rootf+`(i+1)`+"-"+`(i)`+'-coeff.dat'

        print coeff_file
        f=shelve.open(coeff_file)
        c=f['coeff']
        f.close()
        wim.setC(c)
        #IMPORTANT the function getDisplacementAtpoint gives the
        #coordinate transformation to be done to find the
        #coordinates of the gray values in the original image. If
        #we want to apply the transformation to the original image
        #coordinates the sign of the transformation should be
        #changed. (This also happens to the Deformation Table)
        if reverse:
           pointdef[nim-2-i]=-1.0*wim.getDisplacementAtPoint(point+pointdef[nim-1-i])
           pointdefcum[nim-2-i]=pointdefcum[nim-1-i]+pointdef[nim-i-2]
        else:
           pointdef[i+1]=-1.0*wim.getDisplacementAtPoint(point+pointdef[i])
           pointdefcum[i+1]=pointdefcum[i]+pointdef[i+1]

    return pointdefcum
        

# ------------------------------------------------------------------------

def deformation_sequence_with_first(rootf,nim,point):
    """ Takes all the images all the deformations from the deformation tables
    and plots the cumulative deformation.
    """
    
    pointdef=zeros((nim,2),Float)
    img=zeros(imsize_tiff)
    wim=bigregister.WarpableImage(img,imageiscoefs=1)
    for i in range(nim-1):
        coeff_file=rootf+`(i+1)`+"-"+`(0)`+'-coeff.dat'
        f=shelve.open(coeff_file)
        c=f['coeff']
        f.close()
        wim.setC(c)
        #IMPORTANT the function getDisplacementAtpoint gives the
        #coordinate transformation to be done to find the
        #coordinates of the gray values in the original image. If
        #we want to apply the transformation to the original image
        #coordinates the sign of the transformaiton should be
        #changed. (This also happens to the Deformation Table)
        #In this case the deformation is already the cumulative deformation.
        pointdef[i+1]=-1.0*wim.getDisplacementAtPoint(point)
        #print point+pointdef[i], pointdef[i+1]
        
    #myplot_x=bigtools.plot(pointdef[:,0],title='X Cumulative Deformation')
    #myplot_y=bigtools.plot(pointdef[:,1],title='Y Cumulative Deformation')
    #raw_input('Press any key')
        
    return pointdef

# ------------------------------------------------------------------------

def Sequence_DeformationDrawingsAndError(points,nim=nimages, rootf=rootfilename):

    dxl=[]
    dyl=[]
    npoints=points.shape[0]
    f=open((rootf+'cumerr.txt'),'w')
    gx=Gnuplot.Gnuplot()
    gy=Gnuplot.Gnuplot()

    for i in range(npoints):
       point=points[i]
       pointdef1=deformation_sequence(rootf,nim, point)
       #pointdef2=deformation_sequence_with_first(rootf,nim,point)
       #pointdef1=pointdef1
       #pointdef2=pointdef2

       dx1=Gnuplot.Data(arange(nim),pointdef1[:,1], with='lines', title='X Disp P '+`i` )
       #dx2=Gnuplot.Data(arange(nim),pointdef2[:,1], with='linespoints', title='X'' Def P '+`i`)
       #apply(gx.plot,(dx1,dx2))
       apply(gx.plot,(dx1,))
       
       dx1=Gnuplot.Data(arange(nim),10*i+pointdef1[:,1], with='lines', title='X Disp P '+`i` )
       #dx2=Gnuplot.Data(arange(nim),10*i+pointdef2[:,1], with='linespoints', title='X'' Def P '+`i`)
       #dxl=dxl+[dx1,dx2]
       dxl+=[dx1,]
                
       dy1=Gnuplot.Data(arange(nim),pointdef1[:,0], with='lines',title='Y Def P '+`i`)
       #dy2=Gnuplot.Data(arange(nim),pointdef2[:,0], with='linespoints',title='Y'' Def P '+`i`)
       #apply(gy.plot,(dy1,dy2))
       apply(gy.plot,(dy1,))
       dy1=Gnuplot.Data(arange(nim),10*i+pointdef1[:,0], with='lines',title='Y Def P '+`i`)
       #dy2=Gnuplot.Data(arange(nim),10*i+pointdef2[:,0], with='linespoints',title='Y'' Def P '+`i`)


       #dyl=dyl+[dy1,dy2]
       dyl=dyl+[dy1,]
       #apply(gx.plot,(dx1,dx2))
       #apply(gy.plot,(dy1,dy2))

       gx.hardcopy((rootf+'P'+`i`+'gx-real.eps'))
       gy.hardcopy((rootf+'P'+`i`+'gy-real.eps'))
       #f.write(('P '+`i`+' Cumulative error after one cycle (errx, erry) = ('+`abs(pointdef1[nim-1,1]-pointdef2[nim-1,1])`+','+\
    #`abs(pointdef1[nim-1,0]-pointdef2[nim-1,0])`+')\n'))

    f.close()
    
    gx('set size 1,8')
    gy('set size 1,8')
    apply(gx.plot,dxl)
    apply(gy.plot,dxl)
    gx.hardcopy((rootf+'gx-real.eps'))
    gy.hardcopy((rootf+'gy-real.eps'))
    
    raw_input('Press any key')


# ------------------------------------------------------------------------
def test_Sequence_DeformationDrawingsAndError(rootf=rootfilename):
    #points=[[208,174],[162,157],[130,151],[78,177],[93,212],[126,239]] (Real and Model sequence)
    #points=[[96,207],[91,221],[102,137],[239,252],[244,260]] #hipoquinetico 2C-Clip0023
    points=[[122,244],[107,189],[114,144],[118,88],[292,252], [304,200]] #hipoquinetico 2C-Clip0009
    points=array(points, Float)        
    Sequence_DeformationDrawingsAndError(points,nimages,rootf)

# ------------------------------------------------------------------------
# ------------------------------------------------------------------------

def Sequence_ForwRevDrawingsAndError(points, rootfforward,rootfreverse, nim=nimages):

    dxl=[]
    dyl=[]
    npoints=points.shape[0]
    f=open((rootfreverse+'cumerr.txt'),'w')
    gx=Gnuplot.Gnuplot()
    gy=Gnuplot.Gnuplot()

    for i in range(npoints):
       point=points[i]
       pointdef1=deformation_sequence(rootfforward,nim, point)
       pointdef2=deformation_sequence(rootfreverse,nim,point,reverse=1)
       #pointdef1=pointdef1
       #pointdef2=pointdef2

       dx1=Gnuplot.Data(arange(nim),pointdef1[:,1], with='lines', title='X Disp P '+`i` )
       dx2=Gnuplot.Data(arange(nim),pointdef2[:,1], with='linespoints', title='X Disp'' P '+`i`)
       apply(gx.plot,(dx1,dx2))
       dx1=Gnuplot.Data(arange(nim),10*i+pointdef1[:,1], with='lines', title='X Disp P '+`i` )
       dx2=Gnuplot.Data(arange(nim),10*i+pointdef2[:,1], with='linespoints', title='X Disp'' P '+`i`)
       dxl=dxl+[dx1,dx2]

       dy1=Gnuplot.Data(arange(nim),pointdef1[:,0], with='lines',title='Y Disp P '+`i`)
       dy2=Gnuplot.Data(arange(nim),pointdef2[:,0], with='linespoints',title='Y Disp'' P '+`i`)
       apply(gy.plot,(dy1,dy2))
    
       dy1=Gnuplot.Data(arange(nim),10*i+pointdef1[:,0], with='lines',title='Y Disp P '+`i`)
       dy2=Gnuplot.Data(arange(nim),10*i+pointdef2[:,0], with='linespoints',title='Y Disp'' P '+`i`)


       dyl=dyl+[dy1,dy2]
       #apply(gx.plot,(dx1,dx2))
       #apply(gy.plot,(dy1,dy2))

       gx.hardcopy((rootfreverse+'P'+`i`+'gx-real.eps'))
       gy.hardcopy((rootfreverse+'P'+`i`+'gy-real.eps'))
       f.write(('P '+`i`+' Cum Forward Error (cycle) (errx, erry) = ('+`abs(pointdef1[nim-1,1]-pointdef2[nim-1,1])`+','+\
               `abs(pointdef1[nim-1,0]-pointdef2[nim-1,0])`+')\n'))
       f.write(('P '+`i`+' Cum Reverse Error (cycle) (errx, erry) = ('+`abs(pointdef1[0,1]-pointdef2[0,1])`+','+\
               `abs(pointdef1[0,0]-pointdef2[0,0])`+')\n'))

    f.close()
    
    gx('set size 1,8')
    gy('set size 1,8')
    apply(gx.plot,dxl)
    apply(gy.plot,dxl)
    gx.hardcopy((rootfreverse+'gx-real.eps'))
    gy.hardcopy((rootfreverse+'gy-real.eps'))
    
    raw_input('Press any key')


# ------------------------------------------------------------------------
def test_Sequence_ForwRevDrawingsAndError(rootfforward,rootfreverse):
    points=[[208,174],[162,157],[130,151],[78,177],[93,212],[126,239]]
    points=array(points, Float)        
    Sequence_ForwRevDrawingsAndError(points,rootfforward,rootfreverse)

# ------------------------------------------------------------------------

def test_register():
    """ Test the registration algorithm on two sample images """
    cshape=(2,6,6)
    img1=readRawMatlabBytes(imname1,imsize)
    img2=readRawMatlabBytes(imname2,imsize)
    register_pair_and_show(img1,img2,cshape)

    
# ------------------------------------------------------------------------

def test_register_sequence():
    """ Test the registration algorithm on two sample images """
    
    #img1=readRawMatlabBytes(imname1,imsize)
    #img2=readRawMatlabBytes(imname2,imsize)
    #resizing_sequence(rootfilename,nimages,scalefactor)
    #imdirnew="images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6/"
    imdirnew="images_chus/Imn-Chus-4C-Clip0009/results-grid-8x8/"
    imdirrev="images_chus/Imn-Chus-4C-Clip0009/results-grid-8x8-reverse/"
    cshape=(2,8,8)
    tol=1e-3

    #imdirnew="images_chus/Im-2c-Clip0013-masked/results-grid8x8/"
    #rootf=imdirnew+"2C-Clip0013-768x576-"
    rootfforward=imdirnew+"Clip0009-768x576-"
    rootfreverse=imdirrev+"Clip0009-768x576-"
    register_sequence(rootfforward,nimages, cshape, tol)
    register_sequence(rootfreverse,nimages, cshape,tol,reverse=1)
    #register_sequence_with_first(rootf,nimages, cshape, tol)    
    #test_Sequence_ForwRevDrawingsAndError(rootfforward,rootfreverse)

    #cshape=(2,12,12)
    #tol=1e-4
    #imdirnew="images_chus/Im-2c-Clip0013-masked/results-grid-12x12-t1e-4/" 
    #rootf=imdirnew+"2C-Clip0013-768x576-" 
    #register_sequence(rootf,nimages, cshape, tol)
    #register_sequence_with_first(rootf,nimages, cshape, tol)    
    #test_Sequence_DeformationDrawingsAndError(rootfforward)
    

# ------------------------------------------------------------------------

def show_grid():
    imgrid=Image.open(grname)
    matgrid=bigtools.ImageToArray(imgrid) # convert to matrix
    bigtools.view(matgrid)



if __name__=='__main__': # running interactively
    #testTcycle()
    #testMOS()
    #testFullT()
    #test_register_sequence()
    pass
